'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.clearAll = exports.fireOnSequence = exports.dispatchActionWhen = exports.SEQ = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };
/**
 * Created by andrew on 10/13/16.
 */
/**
 * Schedules actions with { meta: { delay: N } } to be delayed by N milliseconds.
 * Makes `dispatch` return a function to cancel the timeout in this case.
 */


var _fluxStandardAction = require('flux-standard-action');

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _istypes = require('istypes');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var symTag = Symbol('SEQ');
var symRegistryId = Symbol('registry');
var symOnce = Symbol('once');

var symNext = Symbol('next');
var symRestart = Symbol('restart');

var symRequired = Symbol('required');
var symEmpty = Symbol('empty');
var symTruthy = Symbol('truthy');
var symFalsey = Symbol('falsey');

var registry = {};
var sequences = [];

function register(fn, effectAction) {

  var unregister = function unregister() {
    var idx = sequences.indexOf(fn);
    if (idx >= 0) {
      return sequences.splice(idx, 1);
    }
    delete registry[id];
  };

  sequences.push(fn);

  var id = Math.random().toString().substr(2);

  fn[symRegistryId] = id;
  registry[id] = {
    unregister: unregister,
    effectAction: effectAction
  };

  return unregister;
}

function unregister(fn) {
  var registryId = fn[symRegistryId];
  if (registryId) {
    if (registry[registryId]) {
      registry[registryId].unregister();
    } else {
      delete registry[registryId];
    }
    delete fn[symRegistryId];
  }
  return fn;
}

function getAction(fn) {
  var registryId = fn[symRegistryId];
  if (registryId) {
    if (registry[registryId]) {
      return registry[registryId].effectAction;
    }
  }
}

function isTagged(input) {
  return !!input[symTag];
}
function tag(input, description) {
  if (_istypes.check.isUndefined(description)) {
    return input[symTag];
  } else {
    input[symTag] = description;
    return input;
  }
}
function tokenToString() {
  return this[symTag];
}
function isOnce(fn) {
  return fn[symOnce];
}
function setOnce(fn) {
  fn[symOnce] = true;
  return fn;
}

/**
 * Internal utility function that wraps result function
 * @param fn
 * @param tokenDescription
 * @param replacement
 * @returns {*}
 */
function tagResultFunction(fn, tokenDescription, replacement) {
  var description = tag(fn);
  if (description) {
    tokenDescription = description.replace(tokenDescription, replacement);
  }

  tag(fn, tokenDescription);

  fn.toString = tokenToString;
  return fn;
}

function restart(tokens) {
  tokens.forEach(function (t) {
    return t(symRestart);
  });
}

function compare(template, obj) {
  if (_istypes.check.isObject(template) && _istypes.check.isObject(obj)) {
    return Object.keys(template).reduce(function (m, key) {
      return m && compare(template[key], obj[key]);
    }, true);
  } else {
    switch (template) {
      case symEmpty:
        return _istypes.check.isUndefined(obj);
      case symRequired:
        return !_istypes.check.isUndefined(obj);
      case symTruthy:
        return !!obj;
      case symFalsey:
        return !obj;
      default:
        return template === obj;
    }
  }
}

var simple = function simple(token) {
  (0, _invariant2.default)(token, 'Token expected for <SIMPLE>');

  var tokenStr = null;

  if (_istypes.check.isArray(token)) {
    return queue(token);
  }

  switch (typeof token === 'undefined' ? 'undefined' : _typeof(token)) {
    case 'string':
      tokenStr = token;
      break;

    case 'function':
      if (isTagged(token)) {
        return token;
      }
      tokenStr = token.toString();
      break;

    case 'object':
      {
        if (_istypes.check.isString(token.type) && Object.keys(token).length == 1) {
          tokenStr = token.type;
        } else {
          return exact(token);
        }
        break;
      }
  }

  (0, _invariant2.default)(tokenStr, 'Token invalid type for <SIMPLE>');

  return tagResultFunction(function (a) {
    return a !== symRestart && (a.type == tokenStr ? symNext : false);
  }, '<SIMPLE>:(' + tokenStr + ')');
};

/**
 * exactObjectSequenceCreator
 * @param token
 * @returns {*}
 */
var exact = function exact(token) {
  (0, _invariant2.default)(token && _istypes.check.isObject(token), 'Token of plain object type expected for <EXACT>');

  return tagResultFunction(function (a) {
    return a !== symRestart && (compare(token, a) ? symNext : false);
  }, '<SIMPLE>:(' + JSON.stringify(token) + ')');
};

var once = function once(token) {
  var normalizedToken = simple(token);
  return tagResultFunction(setOnce(function (a) {
    return normalizedToken(a);
  }), '<ONCE>:(' + normalizedToken.toString() + ')');
};

var timesFn = function timesFn(token, times) {
  var _ref = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  var strict = _ref.strict;

  (0, _invariant2.default)(token, 'Token expected for <TIMES>');

  var normalizedToken = simple(token);
  var tokenDescription = '<TIMES>:(' + normalizedToken.toString() + ' x ' + times + ')';

  var count = 0;
  var result = function result(a) {
    if (a === symRestart) {
      count = 0;
      normalizedToken(symRestart);
      return false;
    }
    var result = normalizedToken(a);
    if (result === true) {
      return true;
    }

    if (result === symNext) {
      count = ++count % times;
      if (count === 0) {
        return symNext;
      } else {
        return true;
      }
    }

    if (strict) {
      count = 0;
      normalizedToken(symRestart);
      return false;
    }
    return true;
  };

  return tagResultFunction(result, tokenDescription);
};

var timesStrict = function timesStrict(token, times) {
  (0, _invariant2.default)(token, 'Token expected for <TIMES_STRICT>');

  return tagResultFunction(timesFn(token, times, { strict: true }), '<TIMES>', '<TIMES_STRICT>');
};

var all = function all(tokens) {
  var _ref2 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref2.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for <ALL>');

  var normalizedTokens = tokens.map(simple);
  var tokenDescription = '<ALL>:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var results = normalizedTokens.map(function (t) {
    return false;
  });
  var length = normalizedTokens.length;

  var result = function result(a) {
    if (a === symRestart) {
      results = normalizedTokens.map(function (t) {
        return false;
      });
      restart(normalizedTokens);
      return false;
    }

    var nextResults = normalizedTokens.map(function (t, k) {
      var result = t(a);

      if (result === symNext) {
        return symNext;
      }
      return results[k];
    }, results);

    var nextResultsReadyCount = nextResults.filter(function (i) {
      return i === symNext;
    }).length;
    if (nextResultsReadyCount === length) {
      results = normalizedTokens.map(function (t) {
        return false;
      });
      restart(normalizedTokens);
      return symNext;
    } else if (strict && nextResultsReadyCount - results.length !== 1) {
      restart(normalizedTokens);
      return false;
    }

    results = nextResults;
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var any = function any(tokens) {
  var _ref3 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref3.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for <ANY>');

  var normalizedTokens = tokens.map(simple);
  var tokenDescription = '<ANY>:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var result = function result(a) {
    if (a === symRestart) {
      restart(normalizedTokens);
      return false;
    }
    if (normalizedTokens.filter(function (t) {
      return t(a) === symNext;
    }).length) {
      restart(normalizedTokens);
      return symNext;
    }
    if (strict) {
      restart(normalizedTokens);
      return false;
    }
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var queue = function queue(tokens) {
  var _ref4 = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var strict = _ref4.strict;

  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for <QUEUE>');

  var normalizedTokens = tokens.map(simple);
  var tokenDescription = '<QUEUE>:(' + normalizedTokens.map(function (t) {
    return t.toString();
  }).join() + ')';

  var length = normalizedTokens.length;
  var count = 0;

  var result = function result(a) {
    if (a === symRestart) {
      count = 0;
      restart(normalizedTokens);
      return false;
    }
    var result = normalizedTokens[count](a);
    if (result === true) {
      return true;
    }

    if (result === symNext) {
      count = ++count % length;

      if (count === 0) {
        restart(normalizedTokens);
        return symNext;
      } else {
        return true;
      }
    }

    if (strict) {
      count = 0;
      restart(normalizedTokens);
      return false;
    }
    return true;
  };
  return tagResultFunction(result, tokenDescription);
};

var queueStrict = function queueStrict(tokens) {
  (0, _invariant2.default)(tokens && tokens.length, 'Tokens expected for <QUEUE_STRICT>');

  return tagResultFunction(queue(tokens, { strict: true }), '<QUEUE>', '<QUEUE_STRICT>');
};

var sequenceApi = {
  simple: simple,
  exact: exact,
  once: once,
  times: timesFn,
  timesStrict: timesStrict,
  all: all,
  any: any,
  queue: queue,
  queueStrict: queueStrict,

  present: symRequired,
  missing: symEmpty,
  truthy: symTruthy,
  falsey: symFalsey,

  /**
   * @deprecated
   */
  SINGLE: simple,
  /**
   * @deprecated
   */
  EXACT: exact,
  /**
   * @deprecated
   */
  ONCE: once,
  /**
   * @deprecated
   */
  TIMES: timesFn,
  /**
   * @deprecated
   */
  TIMES_STRICT: timesStrict,
  /**
   * @deprecated
   */
  ALL: all,
  /**
   * @deprecated
   */
  ANY: any,
  /**
   * @deprecated
   */
  SEQUENCE: queue,
  /**
   * @deprecated
   */
  QUEUE: queue,
  /**
   * @deprecated
   */
  SEQUENCE_STRICT: queueStrict,
  /**
   * @deprecated
   */
  QUEUE_STRICT: queueStrict,

  /**
   * @deprecated
   */
  PRESENT: symRequired,
  /**
   * @deprecated
   */
  MISSING: symEmpty,
  /**
   * @deprecated
   */
  TRUTHY: symTruthy,
  /**
   * @deprecated
   */
  FALSEY: symFalsey

};

/**
 * @deprecated: use 'sequenceBuilderCb(s => ...) 'callback in 'dispatchWhen'
 * @type {{SINGLE: simple, TIMES: timesFn, TIMES_STRICT: timesStrict, ALL: all, ANY: any, SEQUENCE: queue, QUEUE: queue, SEQUENCE_STRICT: queueStrict, QUEUE_STRICT: queueStrict}}
 */
var SEQ = exports.SEQ = sequenceApi;

var dispatchActionWhen = exports.dispatchActionWhen = function dispatchActionWhen(action, sequenceBuilderCb) {

  var sequenceRaw = sequenceBuilderCb(_extends({}, sequenceApi));
  var sequence = simple(sequenceRaw);

  (0, _invariant2.default)(sequence && _istypes.check.isFunction(sequence), 'dispatchActionWhen expects *Sequence Builder* to return compiled sequence (a function)');

  return function (dispatch) {
    return register(sequence, action);
  };
};

/**
 * @deprecated: use dispatchWhen
 * @param action
 * @param sequence
 * @param once
 * @returns {function(*)}
 */
var fireOnSequence = exports.fireOnSequence = function fireOnSequence(action, sequence) {
  return dispatchActionWhen(action, simple(sequence));
};

var clearAll = exports.clearAll = function clearAll() {
  sequences = [];
};

/**
 * Redux middleware to support redux-actions-sequences
 */

exports.default = function (store) {
  return function (next) {
    return function (action) {
      var result = next(action);

      if (sequences.length && (0, _fluxStandardAction.isFSA)(action)) {
        var triggered = [].concat(_toConsumableArray(sequences)).filter(function (s) {
          return s(action) === symNext;
        });
        if (!triggered.length) {
          return;
        }
        var dispatched = triggered.map(function (sequence) {
          sequence(symRestart);

          var action = getAction(sequence);
          var isUnregistered = false;

          var unregisterFn = function unregisterFn() {
            unregister(sequence);
            isUnregistered = true;
          };

          if (_istypes.check.isFunction(action)) {
            store.dispatch(action(unregisterFn));
          } else if (_istypes.check.isString(action)) {
            var seqKey = 'Sequence: ' + sequence.toString() + ' => ' + (action.type || action.toString());
            store.dispatch({
              type: action,
              payload: seqKey,
              meta: {
                unregister: unregisterFn
              }
            });
          } else {
            store.dispatch(action);
          }

          return isUnregistered ? null : isOnce(sequence) ? sequence : null;
        }).filter(function (s) {
          return !!s;
        });

        if (!dispatched.length) {
          return;
        }

        dispatched.forEach(unregister);
      }

      return result;
    };
  };
};